from dataclasses import dataclass
from typing import List, Optional

import numpy as np
from pepeline import color_levels

from reline.static import Node, NodeOptions, ImageFile


@dataclass(frozen=True)
class LevelOptions(NodeOptions):
    low_input: Optional[int] = 0
    high_input: Optional[int] = 255
    low_output: Optional[int] = 0
    high_output: Optional[int] = 255
    gamma: Optional[float] = 1.0


class LevelNode(Node[LevelOptions]):
    def __init__(self, options):
        super().__init__(options)

    def process(self, files: List[ImageFile]) -> List[ImageFile]:
        for file in files:
            file.data = color_levels(
                file.data,
                self.options.low_input,
                self.options.high_input,
                self.options.low_output,
                self.options.high_output,
                self.options.gamma,
            )

        return files

    def single_process(self, file: ImageFile) -> ImageFile:
        file.data = color_levels(
            file.data,
            self.options.low_input,
            self.options.high_input,
            self.options.low_output,
            self.options.high_output,
            self.options.gamma,
        )
        return file

    def video_process(self, file: np.ndarray) -> np.ndarray:
        file = color_levels(
            file,
            self.options.low_input,
            self.options.high_input,
            self.options.low_output,
            self.options.high_output,
            self.options.gamma,
        )
        return file



from __future__ import annotations

from dataclasses import dataclass
from typing import List

from pepeline import save
from reline.static import Node, NodeOptions, ImageFile


@dataclass(frozen=True)
class FileWriterOptions(NodeOptions):
    path: str


class FileWriterNode(Node[FileWriterOptions]):
    def __init__(self, options: FileWriterOptions):
        super().__init__(options)

    def process(self, files: List[ImageFile]):
        if len(files) != 1:
            raise ValueError('Expected single image file')

        return save(files[0].data, self.options.path)

    def single_process(self, file: ImageFile):
        return save(file.data, self.options.path)

    def video_process(self, _):
        raise ValueError('Video scale does not support file write')



from dataclasses import dataclass
from typing import Optional, List, Literal

import numpy as np
import torch.cuda
from resselt import load_from_file
from resr.tiling import MaxTileSize, ExactTileSize, NoTiling, process_tiles
from pepeline import cvt_color, CVTColor
from reline.static import Node, NodeOptions, ImageFile
import logging

Tiler = Literal['exact', 'max', 'no_tiling']
DType = Literal['F32', 'F16', 'BF16']


def empty_cuda_cache():
    torch.cuda.empty_cache()
    torch.cuda.ipc_collect()


@dataclass(frozen=True)
class UpscaleOptions(NodeOptions):
    model: str
    tiler: Tiler
    dtype: Optional[DType] = 'F32'
    exact_tiler_size: Optional[int] = 256
    allow_cpu_upscale: Optional[bool] = False


class UpscaleNode(Node[UpscaleOptions]):
    def __init__(self, options: UpscaleOptions):
        super().__init__(options)

        if not torch.cuda.is_available() and not options.allow_cpu_upscale:
            raise BaseException('CUDA is not available. If you want scale with CPU use `allow_cpu_upscale` option')

        self.model = load_from_file(options.model)
        self.tiler = self._create_tiler()
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        if options.dtype == 'F16':
            self.dtype = torch.half
        elif options.dtype == 'BF16':
            self.dtype = torch.bfloat16
        else:
            self.dtype = torch.float32
        if self.device == 'cuda':
            empty_cuda_cache()

    def _img_ch_to_model_ch(self, img: np.ndarray) -> np.ndarray:
        img_shape = img.shape
        img = img.squeeze()
        if self.model.parameters_info.in_channels == 3:
            if len(img_shape) == 2:
                img = cvt_color(img, CVTColor.Gray2RGB)
        elif self.model.parameters_info.in_channels == 1:
            if len(img_shape) == 3:
                img = cvt_color(img, CVTColor.RGB2Gray_2020)
        else:
            logging.error('model format is not currently supported')
        return img

    def _create_tiler(self):
        match self.options.tiler:
            case 'exact':
                if self.options.exact_tiler_size is None:
                    raise ValueError('Exact tiler requires `exact_tiler_size` param')
                return ExactTileSize(self.options.exact_tiler_size)
            case 'max':
                return MaxTileSize()
            case 'no_tiling':
                return NoTiling()
            case _:
                raise ValueError(f'Unknown tiler option `{self.options.tiler}`')

    def process(self, files: List[ImageFile]) -> List[ImageFile]:
        for file in files:
            img = self._img_ch_to_model_ch(file.data)
            file.data = process_tiles(
                img,
                tiler=self.tiler,
                model=self.model,
                device=self.device,
                dtype=self.dtype,
                scale=self.model.parameters_info.upscale,
                channels=self.model.parameters_info.in_channels,
            )
        return files

    def single_process(self, file: ImageFile) -> ImageFile:
        img = self._img_ch_to_model_ch(file.data)
        file.data = process_tiles(
            img,
            tiler=self.tiler,
            model=self.model,
            device=self.device,
            dtype=self.dtype,
            scale=self.model.parameters_info.upscale,
            channels=self.model.parameters_info.in_channels,
        )
        return file

    def video_process(self, file: np.ndarray) -> np.ndarray:
        img = self._img_ch_to_model_ch(file)
        file = process_tiles(
            img,
            tiler=self.tiler,
            model=self.model,
            device=self.device,
            dtype=self.dtype,
            scale=self.model.parameters_info.upscale,
            channels=self.model.parameters_info.in_channels,
        )
        return file



from __future__ import annotations

import os
from typing import List, Dict

import cv2
import numpy as np
from tqdm import tqdm
from moviepy.editor import VideoFileClip

from ..nodes import INTERNAL_REGISTRY
from ..static import Node
from ..nodes.file_reader import FileReaderNode
from ..nodes.folder_reader import FolderReaderNode
from ..nodes.file_writer import FileWriterNode
from ..nodes.folder_writer import FolderWriterNode


class Pipeline:
    def __init__(self, nodes: List[Node]):
        self.nodes = nodes

    def process(self, with_tqdm: bool = True):
        data = []
        for node in tqdm(self.nodes, desc='Node Processing', disable=not with_tqdm):
            data = node.process(data)
        return data

    def process_linear(self, with_tqdm: bool = True):
        data = []
        nodes_index = 0
        save_index = 0
        while nodes_index < len(self.nodes):
            node = self.nodes[nodes_index]
            if isinstance(node, FileReaderNode | FolderReaderNode):
                data = node.single_process(data)
                for img in tqdm(data, desc='Processing Images', disable=not with_tqdm):
                    if img is None:
                        continue
                    local_node_index = nodes_index + 1
                    for node in self.nodes[local_node_index:]:
                        img = node.single_process(img)
                        local_node_index += 1
                        if isinstance(node, FolderWriterNode | FileWriterNode):
                            save_index = local_node_index - 1
                            break
                nodes_index = save_index
                nodes_index += 1
            else:
                nodes_index += 1
        del data

    def process_frame(self, frame):
        frame = frame.astype(np.float32) / 255
        for node in self.nodes:
            frame = node.video_process(frame)
        if frame.ndim == 2:
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)
        return np.clip(frame * 255, 0, 255).astype(np.uint8)

    def process_video(self, in_folder, out_folder, codec):
        video = VideoFileClip(in_folder).to_RGB()
        processed_video = video.fl_image(self.process_frame)
        audio = video.audio
        final_video = processed_video.set_audio(audio)
        os.makedirs(os.path.dirname(out_folder), exist_ok=True)
        final_video.write_videofile(out_folder, codec=codec)

    @classmethod
    def from_json(cls, data: Dict) -> Pipeline:
        nodes = []

        for item in data:
            node_type = item['type']
            options_data = item['options']

            node_pair = INTERNAL_REGISTRY.get(node_type)
            options = node_pair.options(**options_data)

            node = node_pair.node(options)
            nodes.append(node)

        return Pipeline(nodes)



from __future__ import annotations
from typing import NamedTuple, TypeVar, Dict, Iterator, Type

from ..static import Node, NodeOptions

N = TypeVar('N', bound=Node, covariant=True)
P = TypeVar('P', bound=NodeOptions, covariant=True)


class NodePair(NamedTuple):
    node: Type[N]
    options: Type[P]


class NodePairNotFound(Exception):
    pass


class Registry:
    def __init__(self):
        self.store: Dict[str, NodePair] = {}

    def __contains__(self, name: str):
        return name in self.store

    def __iter__(self) -> Iterator[NodePair]:
        self._iter_keys = iter(self.store)
        return self

    def __next__(self) -> NodePair:
        if self._iter_keys is None:
            raise StopIteration
        try:
            key = next(self._iter_keys)
            return self.store[key]
        except StopIteration:
            self._iter_keys = None
            raise

    def set(self, name: str, node: Type[Node], options: Type[P]) -> Registry:
        self.store[name] = NodePair(node, options)
        return self

    def get(self, name: str) -> NodePair:
        node = self.store[name]
        if not node:
            raise NodePairNotFound
        return node



from __future__ import annotations

import os.path
from dataclasses import dataclass
from typing import List, Optional, Literal

from pepeline import read, ImgFormat, ImgColor

from reline.static import Node, NodeOptions, ImageFile

MODE_MAP = {'rgb': ImgColor.RGB, 'gray': ImgColor.GRAY, 'dynamic': ImgColor.DYNAMIC}

Mode = Literal['rgb', 'gray', 'dynamic']


@dataclass(frozen=True)
class FileReaderOptions(NodeOptions):
    path: str
    mode: Optional[Mode] = 'dynamic'


class FileReaderNode(Node[FileReaderOptions]):
    def __init__(self, options: FileReaderOptions):
        super().__init__(options)
        self.mode = MODE_MAP[options.mode]

    def process(self, _) -> List[ImageFile]:
        basename, _ = os.path.splitext(os.path.basename(self.options.path))
        data = read(self.options.path, color_mode=self.mode, img_format=ImgFormat.F32)

        return [ImageFile(data, basename)]

    def single_process(self, _) -> List[ImageFile]:
        return self.process(0)

    def video_process(self, _):
        raise ValueError('Video scale does not support file read')



import cv2 as cv
import numpy as np

# from chainner_ext import binary_threshold
from pepeline import color_levels


class Canny:
    def __init__(self, canny_type):
        if canny_type == 'invert':
            self.canny_value = 1.0
        else:
            self.canny_value = 0.0

    def run(self, img_float: np.ndarray) -> np.ndarray:
        image = (img_float * 255).astype(np.uint8)
        edges = np.clip(255 - cv.Canny(image, 750, 800, apertureSize=3, L2gradient=False), 0, 1)
        return np.where(edges, img_float, self.canny_value)


class DiapasonWhite:
    def __init__(self, diapason_white: int):
        self.diapason = diapason_white / 255

    def run(self, img_float: np.ndarray) -> np.ndarray:
        image = (img_float * 255).astype(np.uint8)
        median_image = cv.medianBlur(image, 3)
        mask = median_image <= 255 - self.diapason
        return np.where(mask, img_float, 1.0)


class DiapasonBlack:
    def __init__(self, diapason_black: int):
        self.diapason = diapason_black / 255

    def run(self, img_float: np.ndarray) -> np.ndarray:
        black_mask = (img_float > self.diapason).astype(np.float32)
        blur = cv.GaussianBlur(black_mask, (3, 3), 0)
        blur_mask = (blur > 0.9).astype(np.float32)
        return np.where(blur_mask, img_float, 0)


class ColorLevels:
    def __init__(self, low_input, high_input, gamma):
        self.low_input = low_input
        self.high_input = high_input
        self.gamma = gamma

    def run(self, img_float: np.ndarray) -> np.ndarray:
        return color_levels(img_float, self.low_input, self.high_input, 0, 255, self.gamma)
